"use strict";
exports.__esModule = true;
function firstIndexOf(buf, arr) {
    for (var i = 0; i < arr.length; i++) {
        if (buf.equals(arr[i])) {
            return i;
        }
    }
    return -1;
}
var debug = false;
exports.debug = debug;
// Protection against second preimage attacks
// See https://flawed.net.nz/2018/02/21/attacking-merkle-trees-with-a-second-preimage-attack/
var LEAF_PREFIX = Buffer.from('00', 'hex');
exports.LEAF_PREFIX = LEAF_PREFIX;
var BRANCH_PREFIX = Buffer.from('01', 'hex');
exports.BRANCH_PREFIX = BRANCH_PREFIX;
var MerkleTree = /** @class */ (function () {
    function MerkleTree(items, hashFn) {
        var _this = this;
        var leaves = items;
        this.hashFn = hashFn;
        this.hashSizeBytes = hashFn(BRANCH_PREFIX).byteLength;
        // Filter empty
        // leaves = leaves.filter(el => el)
        // check for duplicates
        leaves.filter(function (buf, idx) {
            if (firstIndexOf(buf, leaves) !== idx)
                throw new Error("Duplicate item at " + idx);
        });
        // sort ASC order
        leaves = leaves.sort().reverse();
        // Make sure it's even
        if (leaves.length % 2 == 1) {
            // Some languages (ie Solidity) don't have prepend, so this makes compatible implementations easier.
            leaves = leaves.concat([leaves[leaves.length - 1]]);
            // leaves = [leaves[0], ...leaves]
        }
        // Now hash all.
        leaves = leaves.map(function (leaf) { return _this.hashLeaf(leaf); });
        // And compute tree
        this.layers = this.computeTree(leaves);
    }
    MerkleTree.prototype.root = function () {
        if (this.layers[0].length == 0)
            throw new Error("no leaves in tree");
        return this.layers[this.nLayers - 1][0];
    };
    MerkleTree.prototype.hashLeaf = function (leaf) {
        return hashLeaf(this.hashFn, leaf);
    };
    MerkleTree.prototype.hashBranch = function (left, right) {
        if (left.byteLength != this.hashSizeBytes || right.byteLength != this.hashSizeBytes) {
            throw new Error("branches should be of hash size already");
        }
        return hashBranch(this.hashFn, left, right);
    };
    MerkleTree.prototype.findLeafIndex = function (item) {
        var idx = firstIndexOf(this.hashLeaf(item), this.layers[0]);
        if (idx == -1)
            throw new Error('item not found');
        return idx;
    };
    MerkleTree.prototype.findLeaf = function (item) {
        return this.layers[0][this.findLeafIndex(item)];
    };
    MerkleTree.prototype.generateProof = function (item) {
        var proof = new Array(this.nLayers - 1);
        var idx = this.findLeafIndex(item);
        for (var i = 0; i < this.nLayers - 1; i++) {
            var layer = this.layers[i];
            if (i == this.nLayers - 1) {
                proof[i] = layer[0];
            }
            else {
                var pairIdx = idx % 2 === 0 ? idx + 1 : idx - 1;
                proof[i] = layer[pairIdx];
                idx = Math.floor(idx / 2);
            }
        }
        return proof;
    };
    MerkleTree.prototype.verifyProof = function (proof, leaf) {
        if (proof.length != this.nLayers - 1)
            throw new Error(proof.length + " proof nodes, but only " + this.nLayers + " layers in tree");
        if (firstIndexOf(leaf, this.layers[0]) == -1)
            throw new Error("Leaf doesn't exist in original tree");
        return verifyProof(this.hashFn, proof, this.root(), leaf);
    };
    MerkleTree.prototype.computeTree = function (leaves) {
        // 0th layer is the leaves
        this.nLayers = Math.ceil(Math.log2(leaves.length)) + 1;
        var layers = new Array(this.nLayers);
        for (var i = 0; i < this.nLayers; i++) {
            if (i == 0) {
                layers[i] = leaves;
                continue;
            }
            layers[i] = this.computeLayer(layers[i - 1]);
        }
        return layers;
    };
    MerkleTree.prototype.computeLayer = function (leaves) {
        var nodes = [];
        for (var i = 0; i < leaves.length;) {
            nodes.push(this.hashBranch(leaves[i], leaves[i + 1]));
            i += 2;
        }
        return nodes;
    };
    MerkleTree.prototype.toString = function () {
        var str = "";
        this.layers.map(function (layer, i) {
            str += "Layer " + i + " - \n";
            for (var _i = 0, layer_1 = layer; _i < layer_1.length; _i++) {
                var node = layer_1[_i];
                str += '\t ' + node.toString('hex') + "\n";
            }
        });
        return str;
    };
    return MerkleTree;
}());
exports.MerkleTree = MerkleTree;
function hashLeaf(hashFn, leaf) {
    return hashFn(Buffer.concat([LEAF_PREFIX, leaf]));
}
exports.hashLeaf = hashLeaf;
function hashBranch(hashFn, left, right) {
    return hashFn(Buffer.concat([BRANCH_PREFIX, left, right]));
}
exports.hashBranch = hashBranch;
function verifyProof(hashFn, proof, root, leaf) {
    var node = leaf;
    // node > proof
    // node.compare(proof[0]) == 1
    var dir = node.compare(proof[0]);
    for (var i = 0; i < proof.length; i++) {
        var pairNode = proof[i];
        if (debug) {
            console.log("Verifying layer " + i);
            console.log("\t", node);
            console.log("\t", pairNode);
        }
        if (dir) {
            node = hashBranch(hashFn, pairNode, node);
        }
        else {
            node = hashBranch(hashFn, node, pairNode);
        }
    }
    if (debug) {
        console.log("Verify root");
        console.log('\t', this.root());
        console.log('\t', node);
    }
    return root.equals(node);
}
exports.verifyProof = verifyProof;
//# sourceMappingURL=index.js.map